# coding=UTF-8
"""
Allows you to preserve entire stack frame along with all variables (even
 pickles them).

After this, you can send this report somewhere. At target, it will be unpickled
 in a safe way
(without importing anything extra from environment). You can unpickle
particular variables stored,
but that may involve an import.

Use in such a way:

    try:
        ...
    except WhateverError as e:
        tb = Traceback()
        print(tb.pretty_print())
        # you can now pickle it if you wish to
"""
import inspect
import io
import sys
import traceback
import zlib

import six

try:
    import cPickle as pickle
except ImportError:
    import pickle


class GenerationPolicy(object):
    """
    A policy that manages generating a traceback

    For a "default" policy it's pretty customizable

    Override if need be, and pass the class (or instance) to Traceback
    """

    def __init__(self, enable_pickling=True, compress_at=128 * 1024,
                 repr_length_limit=128 * 1024,
                 compression_level=6):
        """
        :param enable_pickling: bool, whether to enable pickling at all
        :param compress_at: pickles longer than this (bytes) will be compressed
        :param repr_length_limit: maximum length of __repr__. None for no limit.
        :param compression_level: "1" is fastest, "9" is slowest
        """
        assert isinstance(compression_level, int) and 1 <= compression_level <= 9

        self.enable_pickling = enable_pickling
        self.compress_at = compress_at
        self.repr_length_limit = repr_length_limit
        self.compression_level = compression_level

    def should_pickle(self, value):
        """
        Should this value be pickled?

        :param value: value candidate
        :return: bool
        """
        return self.enable_pickling

    def should_compress(self, pickledata):
        """
        Should this pickle undergo compression?
        :param pickledata: bytes, pickle value
        :return: bool
        """
        return len(pickledata) > self.compress_at

    def get_compression_level(self, pickledata):
        """
        What compression level to use to pickle this?
        :param pickledata: bytes, pickle value
        :return: int, 1-9, where "1" is the fastest, and "9" is the slowest,
            but produces best compression
        """
        return self.compression_level

    def process_repr(self, r):
        """
        Process the string obtained from __repr__ing
        :param r: result of a __repr__ on value
        :return: processed result
        """
        if self.repr_length_limit is None:
            return r
        else:
            if len(r) > self.repr_length_limit:
                return r[:self.repr_length_limit] + u'...'
            else:
                return r


class StoredVariable(object):
    """
    Class used to store a variable value. Picklable.

    Attributes are:
        .repr - a text representation obtained using repr
        .typeinfo - a text representation of variable's type
        .pickle - bytes with pickled (optionally processed) value, or None if
            not available
        .pickle_type - what is stored in .pickle?
            None - nothing
            "pickle" - normal Python pickle
            "pickle/gzip" - Python pickle treated with zlib.compress
            "failed" - could not pickle, pickle contains a text with
            human-readable reason

    """
    __slots__ = ('repr', 'type_', 'pickle', 'pickle_type')

    def __init__(self, value, policy):
        """
        If value cannot be pickled, it's repr will be at least preserved

        :param value: any Python value to preserve
        :param policy: policy to use (instance)
        """
        self.repr = repr(value)
        self.type_ = repr(type(value))
        if six.PY2:
            self.repr = six.text_type(self.repr, 'utf8')
            self.type_ = six.text_type(self.type_, 'utf8')

        self.repr = policy.process_repr(self.repr)

        self.pickle = None
        self.pickle_type = None

        if policy.should_pickle(value):
            try:
                self.pickle = pickle.dumps(value, pickle.HIGHEST_PROTOCOL)
                self.pickle_type = 'pickle'
            except BaseException as e:
                # yes, they all can happen!
                self.pickle = repr(e.args)
                self.pickle_type = "failed"
            else:
                if policy.should_compress(self.pickle):
                    try:
                        self.pickle = zlib.compress(
                            self.pickle,
                            policy.get_compression_level(
                                self.pickle))
                        self.pickle_type = "pickle/gzip"
                    except zlib.error:
                        pass  # ok, keep normal

    def load_value(self):
        """
        Return the value that this represents.

        WARNING! This may result in importing things from environment, as
        pickle.loads will be called.

        :return: stored value - if picklable and was pickled
        :raises ValueError: value has failed to be pickled or was never pickled
        """
        if self.pickle_type is None:
            raise ValueError('value was never pickled')
        elif self.pickle_type == 'failed':
            raise ValueError(
                'Value has failed to be pickled, reason is %s' % (self.pickle,))
        elif self.pickle_type == 'pickle/gzip':
            pickle = zlib.decompress(self.pickle)
        elif self.pickle_type == 'pickle':
            pickle = self.pickle

        try:
            return pickle.loads(pickle)
        except pickle.UnpicklingError:
            raise ValueError(
                'object picklable, but cannot load in this environment')


class StackFrame(object):
    """
    Class used to verily preserve stack frames. Picklable.
    """
    __slots__ = ('locals', 'globals', 'name', 'filename', 'lineno')

    def __init__(self, frame, policy):
        """
        :type frame: Python stack frame
        """
        self.name = frame.f_code.co_name
        self.filename = frame.f_code.co_filename
        self.lineno = frame.f_lineno

        self.locals = {}
        for key, value in six.iteritems(frame.f_locals):
            self.locals[key] = StoredVariable(value, policy)

        self.globals = {}
        for key, value in six.iteritems(frame.f_globals):
            self.globals[key] = StoredVariable(value, policy)


class Traceback(object):
    """Class used to preserve exceptions. Picklable."""
    __slots__ = ('formatted_traceback', 'frames')

    def __init__(self, policy=GenerationPolicy):
        """
        To be invoked while processing an exception is in progress

        :param policy: policy for traceback generation
        :raise ValueError: there is no traceback to get info from!
            Is any exception in process?
        """

        if inspect.isclass(policy):
            value_pickling_policy = policy()

        tb = sys.exc_info()[2]

        self.frames = []

        if tb is None:
            raise ValueError('No traceback')
        else:
            while tb.tb_next:
                tb = tb.tb_next

            f = tb.tb_frame
            while f:
                self.frames.append(StackFrame(f, value_pickling_policy))
                f = f.f_back

            self.formatted_traceback = six.text_type(traceback.format_exc())

    def pickle_to(self, stream):
        """Pickle self to target stream"""
        pickle.dump(self, stream, pickle.HIGHEST_PROTOCOL)

    def pickle(self):
        """Returns this instance, pickled"""
        bio = io.BytesIO()
        self.pickle_to(bio)
        return bio.getvalue()

    def pretty_format(self):
        """
        Return a multi-line, pretty-printed representation of all exception
        data.
        :return: text
        """
        bio = io.StringIO()
        self.pretty_print(bio)
        return bio.getvalue()

    def pretty_print(self, output=sys.stderr):
        """
        Pretty-print the exception
        :param output: a file-like object in text mode
        :return: unicode
        """
        k = []
        output.write(self.formatted_traceback)
        output.write(u'\n* Stack trace, innermost first\n')
        for frame in self.frames:
            output.write(u'** %s at %s:%s\n' % (
                frame.name, frame.filename, frame.lineno))
            for name, value in six.iteritems(frame.locals):
                try:
                    output.write(u'*** %s: %s\n' % (name, value.repr))
                except:
                    output.write(u'*** %s: repr unavailable\n' % name)



# coding=UTF-8
"""
Handle signals, like a boss
"""
from __future__ import print_function, absolute_import, division

import logging
import signal
import sys
import time

from satella.coding import typed

logger = logging.getLogger(__name__)

end = False

def __sighandler(a, b):
    global end
    end = True

@typed((list, None))
def hang_until_sig(extra_signals=None):
    """Will hang until this process receives SIGTERM or SIGINT.
    If you pass extra signal IDs (signal.SIG*) with extra_signals,
    then also on those signals this call will release."""
    extra_signals = extra_signals or []
    global end

    # Ascertain what Python are we working on. 2012 PyPy and earlier
    # may be affected by https://bugs.pypy.org/issue1255

    bugged_pypy = False
    try:
        import platform
    except ImportError:
        pass
    else:
        if platform.python_implementation() == 'PyPy':
            try:
                mon, day, year = platform.python_build()[1].split(' ')
                year = int(year)
            except (TypeError, ValueError):
                pass
            else:
                bugged_pypy = year <= 2012

    sleep = lambda: time.sleep(1) if bugged_pypy else signal.pause

    signal.signal(signal.SIGTERM, __sighandler)
    signal.signal(signal.SIGINT, __sighandler)
    for s in extra_signals:
        signal.signal(s, __sighandler)

    while not end:
        try:
            sleep()
        except AttributeError:  # pause() is undefined on Windows
            try:  # we will sleep for small periods of time
                time.sleep(0.5)
            except IOError:  # "Interrupted system call"
                pass



# coding=UTF-8
"""
UNIX things
"""
from __future__ import print_function, absolute_import, division

import logging
import os

from .daemon import daemonize
from .pidlock import AcquirePIDLock
from .signals import hang_until_sig

logger = logging.getLogger(__name__)

__all__ = [
    'daemonize',
    'AcquirePIDLock',
    'hang_until_sig',
    'is_running_as_root',
    'suicide'
]


def is_running_as_root():
    """
    Is this process running as root?

    Checks whether EUID is 0
    :return: bool
    """
    return os.geteuid() == 0


def suicide(kill_entire_pg=True):
    """
    Kill self.

    :param kill_entire_pg: whether to kill entire PG if a session leader
    """
    mypid = os.getpid()
    kill = os.killpg if kill_entire_pg and os.getpgid(0) == mypid else os.kill
    kill(mypid, 9)



# coding=UTF-8
"""
Just useful objects to make your coding nicer every day
"""
from __future__ import print_function, absolute_import, division

from .algos import merge_dicts
from .concurrent import Monitor, RMonitor, CallableGroup, LockedDataset
from .recast_exceptions import rethrow_as, silence_excs
from .structures import TimeBasedHeap, Heap, typednamedtuple, OmniHashableMixin
from .structures import TimeBasedHeap, Heap, typednamedtuple, OmniHashableMixin, Singleton
from .typecheck import typed, Callable, Sequence, \
    TypeVar, Mapping, Iterable, Any, Optional, CallSignature, \
    Number, coerce, Set, Dict, List, Tuple, checked_coerce, for_argument, \
    precondition, PreconditionError
from .decorators import treat_result_with
from .fun_static import static_var

__all__ = [
    'typednamedtuple', 'OmniHashableMixin'
                       'TimeBasedHeap', 'Heap', 'CallableGroup',
    'Monitor', 'RMonitor', 'CallableGroup', 'LockedDataset', 'merge_dicts',
    'typed', 'NewType', 'Callable', 'Sequence', 'coerce'
                                                'TypeVar', 'Mapping', 'Iterable', 'Union', 'Any',
    'Optional',
    'CallSignature', 'Number',
    'Set', 'Dict', 'List', 'Tuple', 'checked_coerce', 'for_argument',
    'precondition', 'PreconditionError',
    'rethrow_as', 'silence_excs',
    'Singleton',
    'treat_result_with',
    'static_var'
]



import logging
import typing
import functools
logger = logging.getLogger(__name__)

__all__ = ['treat_result_with']


def treat_result_with(callable):
    """
    Before this function returns, process it's result with callable

    @treat_result_with(callable)
    def fun(*args, **kwargs):
        ...

    is equivalent to:
    def fun(*args, **kwargs):
        ...

    fun = lambda *args, **kwargs: callable(fun(*args, **kwargs))
    """
    def inner(f):
        @functools.wraps(f)
        def inner2(*args, **kwargs):
            return callable(f(*args, **kwargs))
        return inner2
    return inner



import logging
import typing
import functools

logger = logging.getLogger(__name__)


def static_var(var_name, value):
    """
    Declare a static variable for given function

    Use it like:

    @static_var('counter', 2)
    def count():
        count.counter += 1

    or (syntax valid only on Python 3)

    class MyClass:
        @static_var('counter', 2)
        def count():
            MyClass.counter += 1
    """
    def decorate(func):
        setattr(func, var_name, value)
        return func
    return decorate


# coding=UTF-8
from __future__ import print_function, absolute_import, division

import logging
import os

logger = logging.getLogger(__name__)


class FailedToAcquire(Exception):
    """Failed to acquire the process lock file"""


class LockIsHeld(FailedToAcquire):
    """
    Lock is held by someone

    Has two attributes:
        pid - integer - PID of the holder
        is_alive - bool - whether the holder is an alive process
    """


class AcquirePIDLock(object):
    """
    Acquire a PID lock file.

    Usage:

        with AcquirePIDLock('myservice.pid'):

            .. rest of code ..

    Exiting the context manager deletes the file.

    The constructor doesn't throw, __enter__ does, one of:

    * AcquirePIDLock.FailedToAcquire - base class for errors. Thrown if can't read the file
    * AcquirePIDLock.LockIsHeld - lock is already held. This has two attributes - pid (int), the PID of holder,
                                  and is_alive (bool) - whether the holder is an alive process
    """

    def __init__(self, pid, is_alive):
        super(LockIsHeld, self).__init__()

        self.pid = pid
        self.is_alive = is_alive

    def __init__(self, pid_file, base_dir=u'/var/run', delete_on_dead=False):
        """
        Initialize a PID lock file object

        :param pid_file: rest of path
        :param base_dir: base lock directory
        :param delete_on_dead: delete the lock file if holder is dead, and retry
        """
        self.delete_on_dead = delete_on_dead

        self.path = os.path.join(base_dir, pid_file)

        self.fileno = None

    def _acquire(self):
        """The mechanical process of acquisition"""
        try:
            self.fileno = os.open(self.path, os.O_CREAT | os.O_EXCL)
        except (IOError, OSError):
            try:
                with open(self.path, 'rb') as flock:
                    try:
                        pid = int(flock.read())
                    except ValueError:
                        logger.warning(
                            'PID file found but doesn''t have an int, skipping')
                        return
            except IOError as e:
                raise FailedToAcquire()

            # Is this process alive?
            try:
                os.kill(pid, 0)
            except OSError:  # dead
                raise LockIsHeld(pid, False)
            else:
                raise LockIsHeld(pid, True)

    def __enter__(self):
        try:
            self._acquire()
        except LockIsHeld as e:
            if self.delete_on_dead and (not e.is_alive):
                os.unlink(self.path)
                self._acquire()
            else:
                raise

        self.success = True

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.fileno is not None:
            os.close(self.fileno)
            os.unlink(self.path)



# coding=UTF-8
from __future__ import print_function, absolute_import, division

import numbers
import typing

__all__ = [
    'Callable', 'Sequence', 'Number', 'Mapping', 'Iterable', 'Any',
    'Optional', 'TypeVar', 'List', 'Tuple', 'Dict', 'Set',
    '_NotGiven', '_NoDefault', '_NOP', '_TRUE'
]

Callable = lambda *args: typing.Callable
Sequence = typing.Sequence
Number = numbers.Real
Mapping = typing.Mapping
Iterable = typing.Iterable
Any = typing.Any
Optional = lambda opt: opt + (None,) if isinstance(opt, tuple) else (opt, None)
TypeVar = typing.TypeVar
List = lambda *opt: list
Tuple = lambda *opt: tuple
Dict = lambda *p: dict
Set = lambda *p: set


# Internal tokens - only instances will be
class _NotGiven(object):
    pass


class _NoDefault(object):
    pass


_NOP = lambda x: x
_TRUE = lambda x: True



# coding=UTF-8
from __future__ import print_function, absolute_import, division

import copy
import functools
import heapq
import logging
import operator
import time

import six

from satella.coding.typecheck import typed, Callable, Iterable, Optional, \
    Number

logger = logging.getLogger(__name__)

__all__ = [
    'Heap',
    'TimeBasedHeap',
    'OmniHashableMixin'
]

returns_bool = typed(returns=bool)
returns_iterable = typed(returns=Iterable)


class OmniHashableMixin(object):
    _HASH_FIELDS_TO_USE = []

    def __hash__(self):
        return functools.reduce(operator.xor, (hash(getattr(self, fname)) \
                                               for fname in self._HASH_FIELDS_TO_USE))

    def __eq__(self, other):
        cons = lambda p: [getattr(p, fname) for fname in self._HASH_FIELDS_TO_USE]
        if cons(self) == cons(other):
            return True

        if not isinstance(other, OmniHashableMixin):
            return str(self) == str(other)
        else:
            return False

    def __ne__(self, other):
        return not self.__eq__(other)


def _extras_to_one(fun):
    @functools.wraps(fun)
    def inner(self, a, *args, **kwargs):
        return fun(self, ((a,) + args) if len(args) > 0 else a, **kwargs)

    return inner


class Heap(object):
    """
    Sane heap as object - not like heapq.

    Goes from lowest-to-highest (first popped is smallest).
    Standard Python comparision rules apply.

    Not thread-safe
    """

    def __init__(self, from_list=()):
        self.heap = list(from_list)
        heapq.heapify(self.heap)

    @typed(object, Iterable)
    def push_many(self, items):
        for item in items:
            self.push(item)

    # TODO needs tests
    @_extras_to_one
    def push(self, item):
        """
        Use it like:

            heap.push(3)

        or:

            heap.push(4, myobject)
        """
        heapq.heappush(self.heap, item)

    def __copie(self, op):
        h = Heap()
        h.heap = op(self.heap)
        return h

    def __deepcopy__(self, memo):
        return self.__copie(copy.deepcopy)

    def __copy__(self):
        return self.__copie(copy.copy)

    def __iter__(self):
        return self.heap.__iter__()

    @typed(returns=object)
    def pop(self):
        """
        Return smallest element of the heap.
        :raises IndexError: on empty heap
        """
        return heapq.heappop(self.heap)

    @typed('self', Optional(Callable), Optional(Callable))
    def filtermap(self, filter_fun=None, map_fun=None):
        """
        Get only items that return True when condition(item) is True. Apply a
         transform: item' = item(condition) on
        the rest. Maintain heap invariant.
        """
        heap = filter(filter_fun, self.heap) if filter_fun else self.heap
        heap = map(map_fun, heap) if map_fun else heap
        heap = list(heap) if not isinstance(heap, list) else heap
        self.heap = heap
        heapq.heapify(self.heap)

    @returns_bool
    def __bool__(self):
        """
        Is this empty?
        """
        return len(self.heap) > 0

    @returns_iterable
    def iter_ascending(self):
        """
        Return an iterator returning all elements in this heap sorted ascending.
        State of the heap is not changed
        :return: Iterator
        """
        heap = copy.copy(self.heap)
        while heap:
            yield heapq.heappop(heap)

    @returns_iterable
    def iter_descending(self):
        """
        Return an iterator returning all elements in this heap sorted descending.
        State of the heap is not changed
        :return: Iterator
        """
        return reversed(list(self.iter_ascending()))

    @typed(returns=six.integer_types)
    def __len__(self):
        return len(self.heap)

    def __str__(self):
        return '<satella.coding.Heap: %s elements>' % (len(self, ))

    def __unicode__(self):
        return six.text_type(str(self))

    def __repr__(self):
        return u'<satella.coding.Heap>'

    @returns_bool
    def __contains__(self, item):
        return item in self.heap


class TimeBasedHeap(Heap):
    """
    A heap of items sorted by timestamps.

    It is easy to ask for items, whose timestamps are LOWER than a value, and
    easy to remove them.

    Can be used to implement a scheduling service, ie. store jobs, and each
    interval query
    which of them should be executed. This loses time resolution, but is fast.

    Can use current time with put/pop_less_than.
    Use default_clock_source to pass a callable:

      * time.time
      * monotonic.monotonic

    #notthreadsafe
    """

    def __repr__(self):
        return u'<satella.coding.TimeBasedHeap>'

    @typed(returns=Iterable)
    def items(self):
        """
        Return an iterator, but WITHOUT timestamps (only items), in
        unspecified order
        """
        return (ob for ts, ob in self.heap)

    @typed('self', Optional(Callable((), Number)))
    def __init__(self, default_clock_source=None):
        """
        Initialize an empty heap
        """
        self.default_clock_source = default_clock_source or time.time
        super(TimeBasedHeap, self).__init__(from_list=())

    def put(self, *args):
        """
        Put an item of heap.

        Pass timestamp, item or just an item for default time
        """
        assert len(args) in (1, 2)

        if len(args) == 1:
            timestamp, item = self.default_clock_source(), args[0]
        else:
            timestamp, item = args

        assert timestamp is not None
        self.push((timestamp, item))

    @returns_iterable
    @typed('self', Optional(Number))
    def pop_less_than(self, less=None):
        """
        Return all elements less (sharp inequality) than particular value.

        This changes state of the heap
        :param less: value to compare against
        :return: Iterator
        """

        if less is None:
            less = self.default_clock_source()

        assert less is not None

        while self:
            if self.heap[0][0] >= less:
                return
            yield self.pop()

    def remove(self, item):
        """
        Remove all things equal to item
        """
        self.filtermap(filter_fun=lambda i: i != item)



# coding=UTF-8
from __future__ import print_function, absolute_import, division

"""
Decorator for debug-time typechecking

If you are simultaneously using @typed and @coerce, use them in following order:

    @coerce(...)
    @typed(...)
    def fun(..):
        ...

"""

from .basics import *
from .argparse import *
from .decorators import *

from .decorators import __all__ as __all_decorators__
from .argparse import __all__ as __all_argparse__
from .basics import __all__ as __all_basics__

__all__ = filter(lambda p: not p.startswith('_'),
                 __all_decorators__ + __all_argparse__ + __all_basics__)



# coding=UTF-8
from __future__ import print_function, absolute_import, division
import functools
import warnings
import six

from .corechk import *
from .basics import *

__all__ = [
    'typed',
    'coerce',
    'checked_coerce',
    'for_argument',
    'PreconditionError', 'precondition'
]

def typed(*t_args, **t_kwargs):
    """
    Use like:

        @typed(int, six.text_type)
        def display(times, text):
            ...

    You can also check for return type with kw argument of "returns", ie.

        @typed(int, int, returns=int)
        def sum(a, b):
            return a+b

    Or
        @typed('self', a, b):
        def method(self, a, b):
        ..

    If you specify extra argument - mandatory=True - type will always be
    checked, regardless if debug mode is enabled

    Same rules apply.

    int will automatically include long for checking (Python 3 compatibility)
    If you want to check for None, type (None, )
    None for an argument means "do no checking", (None, ) means "type must be
    NoneType". You can pass tuples or lists to match for multiple types

    :param t_args:
    :param t_kwargs:
    """

    t_args = [(_typeinfo_to_tuple_of_types(x) if x is not None else None)
              for x in t_args]

    t_retarg = t_kwargs.get('returns', None)
    is_mandatory = t_kwargs.get('mandatory', False)

    if t_retarg is not None:
        t_retarg = _typeinfo_to_tuple_of_types(t_retarg)

    def outer(fun):
        if (not __debug__) and (not is_mandatory):
            return fun

        @functools.wraps(fun)
        def inner(*args, **kwargs):
            # add extra 'None' argument if unbound method
            for argument, typedescr in zip(args, t_args):
                if not istype(argument, typedescr):
                    raise TypeError('Got %s, expected %s' % (
                        type(argument), typedescr))

            rt = fun(*args, **kwargs)

            if not istype(rt, t_retarg):
                raise TypeError('Returned %s, expected %s' % (
                    type(rt), t_retarg))

            return rt

        return inner

    return outer


def coerce(*t_args, **t_kwargs):
    """#todo banana banana banana"""
    warnings.warn('Using coerce is considered harmful', DeprecationWarning)

    t_args = [(_typeinfo_to_tuple_of_types(x, operator=_NOP))
              for x in t_args]

    def argify(args):
        return [_do_if_not_type(argument, typedescr) \
                for argument, typedescr in six.moves.zip_longest(args, t_args)]

    t_retarg = t_kwargs.get('returns', None)

    t_retarg = _typeinfo_to_tuple_of_types(t_retarg, operator=_NOP)

    def outer(fun):
        @functools.wraps(fun)
        def inner(*args, **kwargs):
            # add extra 'None' argument if unbound method

            new_args = argify(args)

            rt = fun(*new_args, **kwargs)
            return _do_if_not_type(rt, t_retarg)

        return inner

    return outer


def checked_coerce(*t_args, **t_kwargs):
    """#todo banana banana banana"""
    warnings.warn('Using checked_coerce is considered harmful', DeprecationWarning)

    def ptc(item, pt=list):
        if item is None:
            return None, None
        elif isinstance(item, pt):
            if len(item) == 2:
                return item[0], item[1]
        return item, None

    def sselector(q, z, operator=None, pt=list):
        s = ptc(q, pt=pt)[z]
        if s is None and operator is None:
            return None
        return _typeinfo_to_tuple_of_types(s, operator=operator)

    t_args_t = [sselector(x, 0) for x in t_args]
    t_args_c = [sselector(x, 1, operator=_NOP) for x in t_args]

    def argify(args):
        return [_do_if_not_type(argument, typedescr) \
                for argument, typedescr in six.moves.zip_longest(args, t_args_c)]

    t_retarg = t_kwargs.get('returns', None)
    t_retarg_t = sselector(t_retarg, 0, pt=tuple)
    t_retarg_c = sselector(t_retarg, 1, operator=_NOP, pt=tuple)

    def outer(fun):
        @functools.wraps(fun)
        def inner(*args, **kwargs):
            # add extra 'None' argument if unbound method

            for argument, typedescr in zip(args, t_args_t):
                if not istype(argument, typedescr):
                    raise TypeError('Got %s, expected %s' % (
                        type(argument), typedescr))

            rt = fun(*argify(args), **kwargs)
            if not istype(rt, t_retarg_t):
                raise TypeError('Returned %s, expected %s' % (
                    type(rt), t_retarg_t))

            return _do_if_not_type(rt, t_retarg_c)

        return inner

    return outer


class PreconditionError(ValueError):
    """
    A precondition was not met for the argument
    """


def precondition(*t_ops):
    """
    Check that a precondition happens for given parameter.
    Only positional arguments are supported.

    You can do it like this:

    @precondition(lambda x: x == 1)
    def return_two(x):
        return x*2

    or

    @precondition('x == 1')
    def return_two(x):
        ..

    If None is passed then argument will be always assumed to be True.
    You can use all standard locals in precondition.

    You function call will return a PreconditionError (subclass of
    ValueError) if a precondition fails
    """

    tn_ops = []

    for t_op in t_ops:
        if t_op is None:
            precond = _TRUE
        elif isinstance(t_op, six.string_types):
            q = dict(globals())
            exec('_precond = lambda x: ' + t_op, q)
            precond = q['_precond']
        else:
            precond = t_op

        tn_ops.append(precond)

    from satella.coding.recast_exceptions import rethrow_as

    def outer(fun):
        @functools.wraps(fun)
        def inner(*args, **kwargs):
            assert len(args) >= len(tn_ops), 'More preconditions than positional arguments!'
            with rethrow_as(TypeError, PreconditionError):
                for arg, precond in six.moves.zip_longest(args, tn_ops, fillvalue=_TRUE):
                    print(arg, precond, precond.__doc__)
                    if not precond(arg):
                        raise PreconditionError(
                            'Argument of value %s failed precondition check' % (arg,))
            return fun(*args, **kwargs)

        return inner

    return outer


def for_argument(*t_ops, **t_kwops):
    t_ops = [_NOP if op == 'self' else op for op in t_ops]

    def outer(fun):
        @functools.wraps(fun)
        def inner(*args, **kwargs):
            # add extra 'None' argument if unbound method
            assert len(args) >= len(t_ops)
            return fun(*((_NOP if op is None else op)(arg) for arg, op in
                         six.moves.zip_longest(args, t_ops)),
                       **{k: t_kwops.get(k, _NOP)(v) for k, v in kwargs.items()})

        return inner

    return outer



# coding=UTF-8
from __future__ import print_function, absolute_import, division

import functools

import six

__all__ = [
    'Singleton',
]

if six.PY3:

    # Taken from https://wiki.python.org/moin/PythonDecoratorLibrary
    def singleton(cls):
        """
        Make a singleton out of decorated class.

        Usage:

            @Singleton
            class MyClass(object):
                ...
        """

        cls.__new_old__ = cls.__new__

        @functools.wraps(cls.__new__)
        def singleton_new(cls, *args, **kw):
            it = cls.__dict__.get('__it__')
            if it is not None:
                return it

            cls.__it__ = it = cls.__new_old__(cls, *args, **kw)
            it.__init_old__(*args, **kw)
            return it

        cls.__new__ = singleton_new
        cls.__init_old__ = cls.__init__
        cls.__init__ = object.__init__

        return cls
else:
    class _SingletonWrapper:
        """
        A singleton wrapper class. Its instances would be created
        for each decorated class.
        """

        def __init__(self, cls):
            self.__wrapped__ = cls
            self._instance = None

        def __call__(self, *args, **kwargs):
            """Returns a single instance of decorated class"""
            if self._instance is None:
                self._instance = self.__wrapped__(*args, **kwargs)
            return self._instance


    # taken from https://pypi.python.org/pypi/singleton-decorator/1.0.0
    def singleton(cls):
        """
        A singleton decorator. Returns a wrapper objects. A call on that object
        returns a single instance object of decorated class. Use the __wrapped__
        attribute to access decorated class directly in unit tests.
        """
        return _SingletonWrapper(cls)

Singleton = singleton



# coding=UTF-8
from __future__ import print_function, absolute_import, division

import six

__all__ = [
    '_typeinfo_to_tuple_of_types', 'istype', '_do_if_not_type'
]

from .basics import *


def _typeinfo_to_tuple_of_types(typeinfo, operator=type):
    if typeinfo == 'self':
        return None
    elif typeinfo is None:
        return (operator(None),)
    elif typeinfo == int and six.PY2:
        return six.integer_types
    else:
        if isinstance(typeinfo, (tuple, list)):
            new_tup = []
            for elem in typeinfo:
                new_tup.extend(_typeinfo_to_tuple_of_types(elem))
            return tuple(new_tup)
        else:
            return (typeinfo,)


def istype(var, type_):
    retv = False

    if type_ is None or type_ == 'self':
        retv = True

    elif type(type_) == tuple:
        retv = any(istype(var, subtype) for subtype in type_)

    elif type_ in (Callable, Iterable, Sequence, Mapping):
        if type_ == Callable:
            retv = hasattr(var, '__call__')
        elif type_ == Iterable:
            retv = hasattr(var, '__iter__')
        elif type_ == Sequence:
            retv = hasattr(var, '__iter__') and hasattr(var, '__getattr__') and hasattr(var,
                                                                                        '__len__')
        elif type_ == Mapping:
            retv = hasattr(var, '__getitem__')
    else:
        try:
            if isinstance(var, type_):
                retv = True
        except TypeError:  # must be a typing.* annotation
            retv = type(var) == type_

    return retv


def _do_if_not_type(var, type_, fun='default'):
    if type_ in ((type(None),),) and (fun == 'default'):
        retv = None

    elif type_ in (None, (None,), 'self'):
        retv = var

    elif not istype(var, type_):

        if fun == 'default':
            retv = None if type_[0] == type(None) else type_[0](var)
        else:
            q = fun()
            if isinstance(q, Exception):
                raise q
            retv = q
    else:
        retv = var

    return retv



# coding=UTF-8
from __future__ import print_function, absolute_import, division

__all__ = [
    'typednamedtuple',
    'Heap',
    'TimeBasedHeap',
    'OmniHashableMixin',
    'Singleton'
]

from .singleton import Singleton
from .structures import *
from .typednamedtuple import *



# coding=UTF-8
from __future__ import absolute_import

from .callablegroup import *
from .monitor import *
from .locked_dataset import LockedDataset

__all__ = ['LockedDataset', 'Monitor', 'RMonitor', 'CallableGroup']



# coding=UTF-8
from setuptools import setup, find_packages
from satella import __version__

setup(keywords=['ha', 'high availability', 'scalable', 'scalability', 'server'],
      packages=find_packages(include=['satella', 'satella.*']),
      version=__version__,
      install_requires=[
          "six",
          "monotonic",
          "typing",
      ],
      tests_require=[
          "nose", "mock", "coverage"
      ],
      test_suite='nose.collector',
      python_requires='>=2.7.*,!=3.0.*,!=3.1.*,!=3.2.*,!=3.3.*,!=3.4.*'
      )



import logging
import typing

logger = logging.getLogger(__name__)


class BaseSatellaException(Exception):
    def __init__(self, msg):
        super(BaseSatellaException, self).__init__()
        self.msg = msg

    def __str__(self):
        return 'BaseSatellaException(%s)' % (repr(self.msg), )


class ResourceLocked(BaseSatellaException):
    """Given resource has been already locked"""


class ResourceNotLocked(BaseSatellaException):
    """Locking given resource is needed in order to access it"""



# coding=UTF-8
from __future__ import print_function, absolute_import, division

from .typecheck import typed


def _merge(v1, v2):
    if isinstance(v1, dict) and isinstance(v2, dict):
        for k in v2.keys():
            try:
                v1[k] = _merge(v1[k], v2[k])
            except KeyError:
                v1[k] = v2[k]
        return v1

    if isinstance(v1, list) and isinstance(v2, list):
        v1.extend(v2)
        return v1

    return v2


@typed(dict, dict, returns=dict)
def merge_dicts(first, second):
    return _merge(first, second)
